{	Balls! FKEY  1990 by Jon Wind																	}
{	Version 1.0 on 6/15/90																				}

{	This FKEY can be considered a poor mans screen saver.	 It displays circles in color,			}
{	grays, or b&w, depending on the monitor.  Approximately 60 circles are drawn per			}
{ 	second.  Every 15 seconds the screen will be cleared to black and a new circle size is		}
{	randomly chosen (20-60 pixels).  If the Caps Lock key is down, a random pattern is			}
{	used on color/gray scale monitors; random patterns are always used on b&w monitors.		}
{	It should work correctly even on Macs that have b&w and color monitors.  Press any			}
{	key (except the caps lock and power on keys), click, move the mouse, or insert a disk		}
{	to end the FKEY.																						}

{ 	Thanks to Brad Pettit and his colorfkey for his method of conditional compilation.				}

{	To execute this as a program...																		}
{		1. change the definition of fkey to false														}
{		2. set the project type to application															}
{		3. change the library from drvrruntime.lib to runtime.lib									}
{		4. rebuild the project																			}


{$setc fkey := true}

{$ifc fkey}

unit BallsFKEY;

interface

	procedure main;

implementation

{$elsec}

	program BallsFKEY;

{$endc}

		procedure main;
			const
				bCommandKey = 48;
				bShiftKey = 63;
				bControlKey = 60;
				bOptionKey = 61;
				bCapsLockKey = 62;
				RawMouseGlobal = $82C;
				CreditTime = 180;
				ClsTime = 900;
				Line1 = 'Written by Jon Wind, 6/15/90';
				Line2 = 'MacHack 90';
			var
				ballwidth: Integer;
				iconRect, scrnRect: Rect;
				savePort: GrafPtr;
				oldmouseLoc, oldTicks, lastCls, L: LongInt;
				Credits, hasColor, multiBit, wakeup: Boolean;
				theEvent: EventRecord;
				w: WindowPtr;
				region: RgnHandle;

			function GetLRandom (min, max: Longint): Longint;
	{ return a random number within a given range }
				var
					y: Longint;
			begin
				y := min;
				if min < max then
					repeat
						y := random;
						y := (((y + maxint) * max) div 16383) + min;
					until (y >= min) & (y <= max);
				GetLRandom := y;
			end; { of func GetLRandom }

			function GetKeyDown (index: Integer): Boolean;
	{ return the stae of the desired key - true if down; false if up }
				var
					keys: keymap;
			begin
				GetKeys(keys);
				GetKeyDown := bittst(@keys, index);		{ look at entry within the key map }
			end;

			procedure SetColor (theRect: Rect);
	{ set random color and/or pattern depending on location of Rect }
				var
					deviceHdl: GDHandle;
					useColor: Boolean;
					i: Integer;
					theRGB: RGBColor;
					randPattern: Pattern;
			begin
				usecolor := hasColor;
				if hasColor & multiBit then		{ must check depth of screen - may be B&W }
					begin
						LocalToGlobal(theRect.topLeft);
						LocalToGlobal(theRect.botRight);
						deviceHdl := GetMaxDevice(theRect);		{ get deepest device covered by rect }
						useColor := (deviceHdl <> nil) & (deviceHdl^^.gdPMap^^.pixelsize > 2);
					end;
				if useColor then
					begin
						theRGB.red := GetLRandom(-15000, 25000);
						theRGB.green := GetLRandom(-15000, 25000);
						theRGB.blue := GetLRandom(-15000, 25000);
						RGBForeColor(theRGB);
						if GetKeyDown(bCapsLockKey) or multiBit then	{ use patterns on multi-bit multi-monitor set ups }
							begin
								theRGB.red := GetLRandom(-15000, 25000);
								theRGB.green := GetLRandom(-15000, 25000);
								theRGB.blue := GetLRandom(-15000, 25000);
								RGBBackColor(theRGB);
							end
						else
							PenNormal;
					end
				else
					ForeColor(blackColor);

				if GetKeyDown(bCapsLockKey) or not useColor then 	{ use patterns  }
					begin
						for i := 0 to 7 do
							randPattern[i] := GetLRandom(0, 255);
						PenPat(randPattern);
					end
				else
					PenNormal;
			end;{ of proc SetColor }

			procedure Check4Color (var hasColor, multiBit: Boolean);
	{ hasColor = true if using 4 or more "colors", multiBit = true if colorDevices <> totalDevices }
				var
					deviceHdl: GDHandle;
					theWorld: SysEnvRec;
					totalDevices, colorDevices: Integer;
			begin
				colorDevices := 0;				{ assume no color devices }
				totalDevices := 0;				{ assume no devices }
				if (SysEnvirons(1, theWorld) <> envNotPresent) then	{ SysEnvirons call is available }
					if theWorld.hasColorQD then		{ has Color QuickDraw }
						begin
							deviceHdl := GetDeviceList;
							repeat
								if deviceHdl <> nil then
									begin
										totalDevices := Succ(totalDevices);
										if (deviceHdl^^.gdPMap^^.pixelsize > 1) then	{ 4 or more shades? }
											colorDevices := Succ(colorDevices);
									end;
								deviceHdl := GetNextDevice(deviceHdl);
							until deviceHdl = nil;
						end;
				multiBit := (colorDevices <> totalDevices);
				hasColor := (colorDevices > 0);
			end;{ of proc Check4Color }

			function GetRawMouse: LongInt;
	{ get current mouse location }
				var
					thePtr: ^LongInt;
			begin
				thePtr := Pointer(RawMouseGlobal);
				GetRawMouse := thePtr^;
			end;  { of func GetRawMouse }


 { --------- Main Procedure --------- }
		begin
			GetPort(savePort);					{ save current grafport }
			oldmouseLoc := GetRawMouse;
			Credits := True;						{ I'm going to show a brief credits message }
			Check4Color(hasColor, multiBit);		{ determine whether there's one or more color monitors running in color }

			region := NewRgn;
			CopyRgn(GetGrayRgn, region);		{ get total screen area including menu bar }
			scrnRect := region^^.rgnBBox;
			if (scrnRect.top > 0) then
				scrnRect.top := 0;

			if hascolor then
				w := NewCWindow(nil, scrnRect, '', True, altDBoxProc, Pointer(-1), False, 0)
			else
				w := NewWindow(nil, scrnRect, '', True, altDBoxProc, Pointer(-1), False, 0);

			RectRgn(region, w^.portRect);
			UnionRgn(region, w^.visRgn, w^.visRgn);	{ I'm want to cover everything, including the menu bar }

			SetPort(w);						{ set as current port }
			ClipRect(scrnRect);
			PaintRect(scrnRect);
			lastCls := TickCount;			{ get time of last screen clear }

			ObscureCursor;

			ForeColor(whiteColor);		{ white credits text on black background }
			MoveTo(40, 50);
			DrawString(Line1);
			MoveTo(40, 68);
			DrawString(Line2);

			repeat
				oldTicks := TickCount;		{ store current time }

				if not Credits then
					begin
						if TickCount - LastCls > ClsTime then
							begin
								ForeColor(blackColor);
								PenNormal;
								ballwidth := GetLRandom(20, 60);		{ generate new ball width with each screen clear }
								PaintRect(scrnRect);
								lastCls := TickCount;						{ get time of last screen clear }
							end;
						iconRect.top := GetLRandom(scrnRect.top - BSR(ballwidth, 1), scrnRect.bottom);	{ guarantee clipping on screen top }
						iconRect.left := GetLRandom(scrnRect.left - BSR(ballwidth, 1), scrnRect.right);	{ guarantee clipping on screen left }
						iconRect.right := iconRect.left + ballwidth;
						iconRect.bottom := iconRect.top + ballwidth;
						SetColor(iconRect);		{ send rect to proc that sets color/pattern to use }
						PaintOval(iconRect);
					end
				else if (TickCount - LastCls > CreditTime) then		{ past time when credits should be displayed - only executes once }
					begin
						Credits := False;			{ indicate credits no longer need to be displayed }
						LastCls := 0;				{ set to 0 to force a screen clear }
					end;

				repeat
					wakeup := GetNextEvent(mDownMask + keyDownMask + diskMask, theEvent);
				until wakeup or (TickCount > oldTicks);			{ effectively limits drawing speed to approx. one oval per tick }
			until wakeup or (oldmouseLoc <> GetRawMouse) or GetKeyDown(bOptionKey) or GetKeyDown(bCommandKey) or GetKeyDown(bShiftKey) or GetKeyDown(bControlKey);

			InitCursor;						{ restore cursor - we obscured it earlier }
			DisposeRgn(region);			{ don't need it any more }
			DisposeWindow(w);			{ don't need it any more }
			SetPort(savePort);			{ restore previous grafport }
			DrawMenuBar;					{ fix menubar }
		end;	{ main }


{$ifc fkey = false}

	begin
		main;

{$endc}

	end.