/* [INIT31.c] START OF FILE */

/* [---BEGIN---] FILE HEADER COMMENTS */

/**
 ** ========================================================================
 ** INIT31 Plus v1.0
 ** Written by  Jim Wolff
 ** [INIT31.c] - Implementation INIT 31 mechanism
 ** 1989 by  Apollyonics Software Development, Inc.
 ** P.O. Box 12134
 ** Birmingham, MI  48012-1234
 ** ========================================================================
 **
 **
 ** ================
 ** Revision History
 ** ================
 **  89.09.25			(Jim Wolff)
 **			Initial coding started 
 ** 89.10.21			(James Merkle)
 **			Reviewed code found bug in "Guarentee_System_Heap_Memory"
 ** 90.06.15			(Jim Wolff)
 **				Fixed bugs for MacHack
 **				Killed dean's INIT's
 **
 **
 ** =================
 ** OTHER INFORMATION
 ** =================
 **	Palitino 10, Tabs = 14
 **
 **	INIT31 Installer is the installation software used to install INIT-31+
 **
 **	INIT-31 is the piece of code that resides in the "System" file that causes the files with types 'INIT',
 **	'cdev' and 'RDEV' in the "System Folder" to be executed on startup.
 **
 **	INIT-31+ is a INIT-31 replacement that, in addition to doing what INIT-31 does, has these features:
 **		1. Displays information about every INIT on startup
 **		2. Keeps a info file on every INIT on start (future).
 **		3. Searches alternate directories looking for INITs.  This is controlled by the accompaning
 **			Control Panael document (cdev, future.)
 **		4. It is primarily a complete disassembly of Apple's INIT-31 mechanism.
 **
 **
 **
 **	BEWARE: This will break with System 7.0
 **
 **/

/* [----END----] FILE HEADER COMMENTS */

#include <SetUpA4.h>

#define QUICK_DRAW_GLOBALS		-4

#define DefltStack		0x322

#define ioNamePtr		18
#define ioVRefNum		22
#define ioFVersNum		26
#define ioFDirIndex		28
#define ioFlFndrInfo		32
#define fdType		0
#define fdFlags		8

#define BkLim		0

Handle		handle_To_Ourself;
Ptr		saved_A4;
Ptr		saved_Return_Address;		/* HACK to make it work with NEW Init picker */

/* NOTE: Make the rectangle based on where the "Welcome" dialog is.  It is different on different size screens */
Rect		name_Rect = {164, 256, 184, 476 };


char		loading_Message[] = "\pLoading: ";
/* char		loading_Message[] = "\pMacsBug!: "; */

ParamBlockRec	file_Parameter_Block;
Str255		file_Name, last_File_Name;
GrafPtr		saved_Port, drawing_Port_Ptr;
GrafPort		drawing_Port;

void Setup_Drawing_Rectangle(void);
void Copy_Pascal_String(char *source_String, char *destination_String);
Boolean Compare_Pascal_Strings(char *source_String, char *destination_String);

main()
{
	asm {
		movem.l		d0-d7/a0-a6,-(sp)		;Saved until very end of main
	}
	
	RememberA0();
	SetUpA4();

	asm
	{
		move.l	(sp),saved_A4
		_RecoverHandle	SYS			/* a0 already points to resource */
		move.l	a0,handle_To_Ourself
		
		tst.w	ROM85				;Is it a 64k ROM?
		bpl.s	@Make_Stack			;No, take the branch
		
		move.w	(sp),d0				;d0.w = Word at sp
		subq.w	#1,d0				;d0.w = 1?
		bne.s	@Make_Stack			;No, take the branch
		move.w	MemTop,d0			;d0.w = Low word of MemTop
		cmpi.w	#2,d0				;d0.w = 2?
		beq.s	@Make_Stack			;Yes, take the branch
		
		move.l	MemTop,d0			;d0.l = MemTop
		lsr.l	#2,d0				;d0.l = MemTop div 4
		add.l	d0,d0				;d0.l = d0.l + d0.l
		movea.l	d0,a1				;a1 = Pointer to middle of memory
		moveq	#1,d0				;d0.l = 1
		swap	d0				;d0.l = 65536
		movea.l	d0,a0				;a0 = Pointer to 64k
		move.l	#0x400,d0			;d0.l = 1k
		move.l	a1,d1				;d1 = Pointer to middle of memory
		add.l	d0,d1				;d1 = MemTop/2 + 1k
		cmp.l	BufPtr,d1			;Did we run out of static allocation space?
		bhi.s	@Make_Stack			;Yes, take the branch
		
		move.l	d0,d1				;d1.l = 1k
		_BlockMove				;(A0/srcPtr,A1/destPtr:Ptr; D0/byteCount:Size)
							;a0 = Pointer to 64k, a1 = Pointer to middle of memory
		move.l	a1,d0				;d0.l = Pointer to middle of memory
		subi.l	#0x10000,d0			;d0.l = d0.l - 64k
		add.l	d0,(sp)				;Add amount left to stack
		movea.l	a1,a2				;d2 = Pointer to middle of memory
		suba.l	d1,a2				;a2 = a2 - 1k;
		lea	400(a2),a5			;a5 = a2 + 400 bytes
		pea	QUICK_DRAW_GLOBALS(a5)	;Globals start here
		_InitGraf				;(globPtr:Ptr)
		move.l	(sp)+,d0				;d0.l = result
		exg	sp,a2				;Exchange
		move.l	d0,-(sp)				;Put result back on stack?
		
@Make_Stack
		movea.l	sp,a0				;a0 = Pointer to stack
		suba.l	DefltStack,a0			;a0 = (a0 - default stack space) (24k on Mac II)
		_SetApplLimit				;(A0/zoneLimit:Ptr)
		btst	#0xE,ROM85			;Running on a Mac II?
		bne.s	@Mac_II_Skip			;No, take the branch
		pea	QUICK_DRAW_GLOBALS(a5)	;Our quickdraw globals
		_InitGraf				;(globalPtr:Ptr)
		_InitPalettes
		
@Mac_II_Skip
		movem.l	d0-d7/a0-a6,-(sp)		;Saved until #0
		moveq	#-1,d4				;d4.l = -1
		move.l	#0x4000,d7			;d7.l = 16k
		bsr	@Guarentee_System_Heap_Memory
		moveq	#0,d3
		
		movem.l	d0-d7/a0-a6,-(sp)		;Saved for next few lines until restore (#1)
	} /* asm */
	
		Setup_Drawing_Rectangle();
		Copy_Pascal_String("\p", (char *)last_File_Name);
		
	asm
	{
@Restore_And_Try_Next
		movem.l	(sp)+,d0-d7/a0-a6		;Restored from few lines above (#1)

@Try_Next_File
		lea	file_Parameter_Block,a0		;a0 = Pointer to file parameter block
		addq.w	#1,d3				;d3.w = Index for file in directory
		pea	file_Name			;Push address of file_Name
		move.l	(sp)+,ioNamePtr(a0)		;load in into the ioNamePtr field
		clr.w	ioVRefNum(a0)			;Current volume = 0
		clr.w	ioFVersNum(a0)			;ioFVersNum = 0
		move.w	d3,ioFDirIndex(a0)		;ioFDirIndex = Index to file counter
		_PBGetFInfo				;(A0|IOPB:ParamBlockRec):D0\OSErr
		bne	@No_More_Files			;Branch if error
		move.l	ioFlFndrInfo+fdType(a0),d0	;d0.l = File type
		cmpi.l	#'INIT',d0			;Was it an 'INIT'?
		beq.s	@Valid_INIT_File		;Yes, take the branch
		cmpi.l	#'cdev',d0			;Was it a 'cdev'?
		beq.s	@Valid_INIT_File		;Yes, take the branch
		cmpi.l	#'RDEV',d0			;Was it a 'RDEV'?
		bne	@Try_Next_File			;No, take the branch
		
@Valid_INIT_File
		
		movem.l	d0-d7/a0-a6,-(sp)	;Saved for C calls that follow (#2)
		
	} /* asm */
	
		if (Compare_Pascal_Strings((char *)last_File_Name, (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pVolumeDock", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pEqualizer", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pwheel", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pminiFinder", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pThis", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\pPickEnvirons", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
			
		if (Compare_Pascal_Strings("\p!MultiFinder", (char *)file_Name) == TRUE)
			goto Restore_And_Try_Next;
		
	asm
	{
		movem.l	(sp)+,d0-d7/a0-a6		;Restore from C calls (#2)

		btst	#7,ioFlFndrInfo+fdFlags+1(a0)
		bne	@Try_Next_File
		move.w	ioFlFndrInfo+fdFlags(a0),d0
		andi.w	#0x4000,d0
		btst	#0xE,ioFlFndrInfo+fdFlags(a0)
		bne	@Try_Next_File
		bsr	@Open_File
		tst.w	ResErr
		bne	@Try_Next_File
		
		movea.l	TopMapHndl,a0
		movea.l	(a0),a0
		move.l	16(a0),d5
		clr.l	16(a0)
		clr.l	-(sp)				;Room for handle to resource
		move.l	#'sysz',-(sp)			;Going to get a 'sysz' resouce
		clr.w	-(sp)				;With ID = 0
		_GetResource				;(theType:ResType; ID:INTEGER):Handle
		movea.l	TopMapHndl,a0
		movea.l	(a0),a0
		move.l	d5,16(A0)
		move.l	(sp)+,d0				;d0.l = Handle to 'sysz' resource (if any)
		beq.s	@Skip_sysz			;Branch if no 'sysz' resource (it was NIL)
		movea.l	d0,a0				;a0 = Handle to 'sysz', ID = 0 resource
		movea.l	(a0),a0				;a0 = Pointer to 'sysz' resouce
		move.l	(a0),d7				;d7.l = Size of system heap requested from 'sysz'
		bsr	@Guarentee_System_Heap_Memory
		
@Skip_sysz
		moveq	#1,d6				;d6.l = 1 (our index counter)
		
@Try_Next_INIT_In_File
		move.l	#0x4000,d7			;d7.l = 16k
		bsr	@Guarentee_System_Heap_Memory
		movea.l	TopMapHndl,a0
		movea.l	(a0),a0
		move.l	16(a0),d5
		clr.l	16(A0)
		clr.l	-(sp)				;Room for handle to resource
		move.l	#'INIT',-(sp)			;Push type of resouce to read 'INIT'
		move.w	d6,-(sp)				;Push ID of 'INIT' resource to read
		_GetIndResource				;(theType:ResType; index:INTEGER):Handle
		movea.l	TopMapHndl,a0
		movea.l	(a0),a0
		move.l	d5,16(a0)
		move.l	(sp)+,d0				;d0.l = Handle to 'INIT' if any
		beq.s	@INIT_Not_Read		;Branch if no 'INIT' resource (it was NIL)
		tst.w	ResErr				;Was there a resouce error?
		bne.s	@INIT_Not_Read		;Yes, take the branch
		movea.l	d0,a0				;a0 = Handle to 'INIT' resource, index d6.w
		movea.l	(a0),a1				;a1 = Pointer to 'INIT' resource
		
		movem.l	d0-d7/a0-a6,-(sp)		;Save accross calls to INIT (#3)
		movea.l	saved_A4,a4
		
		jsr	(a1)				;Call the INIT
		
		movem.l	(sp)+,d0-d7/a0-a6		;Restore from call of INIT (#3)
		
		addq	#1,d6				;Next index to try
		bra	@Try_Next_INIT_In_File		;Try the next init in the file
		
@INIT_Not_Read
		cmpi.w	#1,d6				;Did we execute ANY 'INIT's in this file?
		bne.s	@Cleanup_And_Try_Next		;Yes, take the branch
		lea	file_Parameter_Block,a0
		bset	#7,ioFlFndrInfo+fdFlags+1(a0)
		_PBSetFInfo				;(A0|IOPB:ParamBlockRec):D0\OSErr

@Cleanup_And_Try_Next
		bsr	@Close_File
		bra	@Try_Next_File
		
@No_More_Files			
		movem.l	(sp)+,d0-d7/a0-a6	;Restore from #0
		
		move.l	handle_To_Ourself,a0		;Recover handle to ourself ('INIT', ID = 31)
		move.l	a0,d7				;Put in d7.l for history sake...
		_DisposHandle			;(A0/h:Handle)
		bra	@Exit_INIT_31

@Open_File
		clr.b	ResLoad
		clr.w	-(sp)
		pea	file_Name
		_OpenResFile				;(fileName:Str255):RefNum
		move.w	(sp)+,d4
		st	ResLoad
	} /* asm */
	
	if (Compare_Pascal_Strings((char *)last_File_Name, (char *)file_Name) == FALSE)
	{
		Copy_Pascal_String((char*)file_Name, (char *)last_File_Name);

		asm
		{
			pea	QUICK_DRAW_GLOBALS(a5)
			_InitGraf				;(globPtr:Ptr)
		} /* asm */
		
		GetPort(&saved_Port);
		OpenPort(&drawing_Port);
		EraseRect(&name_Rect);
		FrameRect(&name_Rect);
		MoveTo(name_Rect.left, name_Rect.bottom);
		LineTo(name_Rect.right, name_Rect.bottom);
		MoveTo(name_Rect.right, name_Rect.top);
		LineTo(name_Rect.right, name_Rect.bottom);
		MoveTo((name_Rect.left+2), (name_Rect.top + 12));
		DrawString(loading_Message);
		DrawString(file_Name);
		ClosePort(&drawing_Port);
		SetPort(saved_Port);
		
	} /* if (Compare_Pascal_Strings((char *)last_File_Name, (char *)file_Name) == FALSE) */
		
	
	asm
	{
		rts

@Close_File
		move.w	d4,-(sp)
		_CloseResFile				;(refNum:RefNum)
		rts

@Guarentee_System_Heap_Memory
		move.l	d7,d0				;d0.l =16k
		_CompactMem	SYS			;(D0/cbNeeded:Size):D0\Size
		sub.l	d7,d0				;Is (d0 > d7)?
		bge.s	@lae_2				;Yes, take the branch
		movea.l	SysZone,a0			;a0 = Pointer to system heap zone
		movea.l	BkLim(a0),a0			;a0 = Pointer to end of system heap zone
		suba.l	d0,a0				;a0 = Pointer to end + amount needed
		adda.w	#128,a0				;Plus 128 bytes extra
		tst.w	d4				;d4?
		bpl.s	@lae_1				;If 0 or >?
		_SetApplBase				;(A0/startPtr:Ptr)
		_InitApplZone
		bra.s	@Guarentee_System_Heap_Memory	;!!!!! ERROR !!!!! either loop forever or crash!
@lae_1
		move.l	a0,-(sp)				;a0 = Pointer to top of stack
		bsr	@Close_File			;Close current file
		move.l	(sp)+,a0				;Restore pointer
		_SetApplBase				;(A0/startPtr:Ptr)
		_InitApplZone
		bsr	@Open_File
		bra.s	@Guarentee_System_Heap_Memory
@lae_2
		rts
	} /* asm */
	
Exit_INIT_31:

	RestoreA4();
	
	asm {
		movem.l		(sp)+,d0-d7/a0-a6		;Restore from very start of main
	}
	
} /* main() */
/*************************************************************************/


/*************************************************************************/
void Setup_Drawing_Rectangle(void)
{
	GrafPort		window_Manager_Port;
	Rect		screen_Rect;
	Str255		message_String;
	
	asm
	{
		lea	screen_Rect,a0
		lea	-120(a5),a1			; -120
		move.l	(a1)+,(a0)+
		move.l	(a1)+,(a0)+
	}
	
	name_Rect.left = (((screen_Rect.right -  screen_Rect.left) / 2) - 64);
	name_Rect.right = ( name_Rect.left + 230);

	name_Rect.top  = ((((screen_Rect.bottom -  screen_Rect.top)*3) / 16) + 70);
	name_Rect.bottom = (name_Rect.top + 18);
	
	
} /* Setup_Drawing_Rectangle */
/*************************************************************************/


/*************************************************************************/
void Copy_Pascal_String(char *source_String, char *destination_String)
{
	unsigned char	length_Byte;
	
	length_Byte = (source_String[0] + 1);
	
	BlockMove(source_String, destination_String, length_Byte);
} /* Copy_Pascal_String */
/*************************************************************************/


/*************************************************************************/
Boolean Compare_Pascal_Strings(char *source_String, char *destination_String)
{
	unsigned char	length_Byte;
	int		count;
/* END VAR */

	if (source_String[0] != destination_String[0])
		return(FALSE);
	
	length_Byte = (source_String[0] + 1);

	for ((count = 1); (count < length_Byte); count++)
		if (source_String[count] != destination_String[count])
			return(FALSE);
	
	return(TRUE);
} /* Boolean Compare_Pascal_Strings(char *source_String, char *destination_String) */
/*************************************************************************/


/* [INIT31.c] END OF FILE */