/************************************************************************************/
/* 									Low Level MIDI Interface Driver								*/
/************************************************************************************/
/* Version: 1.0			Written by Chris Marshall September 22nd, 1989					*/
/* 										1989 by Chris Marshall										*/
/* Written using the THINK C version 4.0 development system.								*/
/************************************************************************************/
/* These are declarations for the MIDI driver that can be used by the application.	*/
/************************************************************************************/

/************************************************************************************/
/* These are the default names of the low-level drivers. - DON'T CHANGE!				*/
/************************************************************************************/

#define	PORTANAME	"\p.MIDI Port A"				/* Port A driver							*/
#define	PORTBNAME	"\p.MIDI Port B"				/* Port B driver							*/

#define	DEFLEN		2048L		/* Default buffer length (in bytes).						*/

/************************************************************************************/
/* These are the port definition flags.															*/
/************************************************************************************/

#define	PORTA			1		/* These are flags to indicate the affected port.			*/
#define	PORTB			0		/* Port A is the modem port, port B is the printer port.	*/

/************************************************************************************/
/* Clock setting indexes																				*/
/************************************************************************************/

#define	CL31KHZ		0			/* 31.25 KHz clock source frequency							*/
#define	CL500KHZ		1			/* 500 KHz															*/
#define	CL1MHZ		2			/* 1 MHz																*/
#define	CL2MHZ		3			/* 2 MHz																*/

/************************************************************************************/
/* These are the control call codes.																*/
/************************************************************************************/

#define	KILLIO		1			/* Standard kill I/O code.										*/
#define	GOODBYE		goodBye	/* Standard "goodbye kiss"										*/

#define	RESETALL		10			/* Reset the driver, and flush the buffers.				*/
#define	FLUSHBUFF	20			/* Flush the buffers												*/
#define	CLOCKRATE	40			/* Change the clock source divisor							*/

/************************************************************************************/
/* NOTE: This command uses memory manager routines.  Do not - repeat: DO NOT call	*/
/* the driver with this command asynchronously or during an interrupt!					*/
/************************************************************************************/

#define	RESIZE		30			/* Resize the buffers											*/

/************************************************************************************/
/* These are status call codes for register checks.											*/
/************************************************************************************/

#define	CHECKREG		10			/* Check SCC registers											*/
#define	STATREC		20			/* Return the status record									*/

/************************************************************************************/
/* This is the structure cast into the csParam field when a status call is made.		*/
/************************************************************************************/

typedef struct MyParms			/* This is the status return's structure.					*/
	{
	int	myClock;					/* The clock index												*/
	long	inBufferSize;			/* The input buffer's size										*/
	long	outBufferSize;			/* The output buffer's size									*/
	long	inBufferChar;			/* The byte count of the input buffer						*/
	long	outBufferChar;			/* The byte count of the output buffer						*/
	int	driverErr;				/* Any existing error conditions								*/
	} MyParms;
	
/************************************************************************************/
