/************************************************************************************/
/* 									Low Level MIDI Interface Driver								*/
/************************************************************************************/
/* Version: 1.0			Written by Chris Marshall September 22nd, 1989					*/
/* 										1989 by Chris Marshall										*/
/* Written using the THINK C version 4.0 development system.								*/
/************************************************************************************/
/* This file contains all of the nasty little assembly language routines used by 	*/
/* the driver.  They're really not so bad!														*/
/************************************************************************************/

#include "PortIO.h"				/* These are all of the constants used by the driver.	*/
#include "MIDI.h"					/* Get the exported constants and declarations			*/
#include "Errors.h"

/************************************************************************************/
/* This nasty little bit of code will "punch a hole" in itself, and will return that*/
/* location to be used as the programmer sees fit.  This allows independent code to */
/* access data without relying on A4, A5 or CurrentA5.  The code MUST be locked and */
/* available during the entire time that this location is being used!               */
/************************************************************************************/

MyDataPtr *MyStash ()
	{
	asm
		{
   		MOVE.L	A0,-(SP)		/* Save the register												*/
			   		
			MOVEA.L	#0,A0			/* Clear the register for the hell of it					*/
			BRA.S		@Go_On		/* Skip around the data 										*/
			DC.L		0				/* The "hole":  This is where our data goes.				*/
@Go_On	LEA.L		(-6,PC),A0	/* Put the "hole"'s location into A0 						*/
			MOVE.L	A0,D0			/* Put the "hole" into the function return 				*/
      	
   		MOVE.L	(SP)+,A0					/* Restore the register								*/
		};
	};
   
/************************************************************************************/
/* This is a "generic" SCC control register writing routine.                        */
/************************************************************************************/

int WriteCtlReg (regNum, data)
unsigned char regNum,      /* The control register number                           */
              data;        /* The data to put into the register                     */
   {
   asm                     			/* Inline assembly makes this a little easier	*/
      {
   	MOVEM.L	A0-A1/D0,-(SP)			/* Save the registers									*/
			   		
      JSR		MyStash					/* Get the private data									*/
      
      MOVEA.L	D0,A1						/* Move pointer's address to address register	*/
      MOVEA.L	(A1),A1					/* Need to do it twice for generic 68000 code	*/
	
      MOVEA.L  SCCWRITE(A1),A0		/* Place register 0 address into address reg.	*/
		TST.B		regNum					/* Check to see if it's register 0					*/
		BEQ		@Skip						/* No need to address it if so						*/
		
      MOVE.B   regNum,(A0) 			/* Select the register to be addressed				*/
      
      /******************************************************/
      /* The SCC requires 4 clock cycles to ensure that the	*/
      /* data is stable.  This is probably because it's a	*/
      /* little "schitzo" in a timing sense, as it has a		*/
      /* serial clock and a system clock, and will need to	*/
      /* ensure that data goes from one world to the other	*/
      /* properly. (Uh, 4 system clock cycles, that is)		*/
      /******************************************************/
      
      MOVE.L   (SP),(SP)   			/* This kills settling time  							*/

@Skip	MOVE.B   data,(A0) 				/* Place the data byte in the register				*/
      MOVE.L   (SP),(SP)   			/* Dum-de-dum dummmmm...								*/
      	
   	MOVEM.L	(SP)+,A0-A1/D0			/* Restore the registers								*/
      };
   };

/************************************************************************************/
/* This routine will read the value of a control register, and return it.           */
/************************************************************************************/

unsigned char ReadCtlReg (regNum)
unsigned char regNum;      			/* The control register number						*/
   {
   asm                     			/* Inline assembly makes this a little easier	*/
      {
   	MOVEM.L	A0-A1,-(SP)				/* Save the registers									*/
			   		
      JSR		MyStash					/* Get the private data									*/
      
      MOVEA.L	D0,A1						/* Move pointer's address to address register	*/
      MOVEA.L	(A1),A1					/* Need to do it twice for generic 68000 code	*/
		
		CLR.L		D0							/* Clear the function return completely			*/

      MOVEA.L  SCCWRITE(A1),A0		/* Place register 0 address into address reg.	*/
		TST.B		regNum					/* Check to see if it's register 0					*/
		BEQ		@Skip						/* No need to address it if so						*/

      MOVE.B	regNum,(A0) 			/* Select the register to be addressed				*/
      MOVE.L   (SP),(SP)   			/* This kills settling time  							*/

@Skip MOVEA.L  SCCREAD(A1),A0			/* Address the read register							*/
      MOVE.L   (SP),(SP)   			/* Dum-de-dum dummmmm...								*/
      MOVE.B   (A0),D0     			/* Place the data into the function return 		*/
      	
   	MOVEM.L	(SP)+,A0-A1				/* Restore the registers								*/
      };
   };

/************************************************************************************/
/* This is the interrupt called routine that enters a byte from the SCC.				*/
/************************************************************************************/

pascal void GetOne ()
	{
   asm
   	{
   		MOVE.W	SR,-(SP)					/* Save the previous status condition			*/
			ORI.W		#0x0300,SR				/* Step the interrupt priority up to 3			*/
   		MOVEM.L	D0-D1/A2-A5,-(SP)		/* Save the registers								*/

   		JSR		MyStash					/* Get the address of the private data			*/
			
   		MOVE.L	D0,A4						/* Address the data structure						*/
   		MOVE.L	(A4),A4					/* Takes two steps for generic 68000 code		*/

		/******************************************************************************/
		/* The driver needs to make sure that it's not going too fast for itself.  	*/
		/* This routine checks the SCC's read register 0 to make sure we have a byte	*/
		/* before we try to input it.																	*/
		/******************************************************************************/
		
@SCCChk	MOVEA.L	SCCWRITE(A4),A1		/* Address port write control register	 		*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			MOVE.B	#0x00,(A1)				/* Address register 0								*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			MOVEA.L	SCCREAD(A4),A0			/* Address port read control register 0 		*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			CLR.L		D1							/* Just for s's and g's, we'll clear D1		*/
			MOVE.B	(A0),D1					/* Put the status byte into D1					*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			BTST		#0x00,D1					/* See if the RX buffer contains anything 	*/
			BEQ		@GitGone					/* Don't do it if the buffer's MT				*/
			
   		MOVEA.L	INBUFF(A4),A5			/* Address the input buffer 						*/
   		MOVEA.L	FILL(A5),A2				/* Address the fill pointer						*/
   	
      	MOVEA.L  SCDREAD(A4),A0			/* Place register address into address reg	*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/

@SCCRead MOVE.B	(A0),(A2)+				/* Put the byte into the buffer					*/

   		MOVE.L	COUNT(A5),D0			/* Get the byte count								*/
   		MOVE.L	SIZE(A5),D1				/* Get the allocated length						*/
   		LEA.L		BUFF(A5),A3				/* Get the beginning of the buffer				*/
   		ADDA.L	D1,A3						/* Point to the end of the buffer				*/
   		SUBA.L	#1,A3						/* Take 1 off to include first byte				*/
   		CMPA.L	A2,A3						/* See if the buffer is at the end				*/
   		BPL		@Skip1					/* Don't worry if we're not at the end			*/
   		
   		LEA.L		BUFF(A5),A2				/* Reset the fill pointer to the beginning	*/
   		
@Skip1	MOVE.L	A2,FILL(A5)				/* Put the modified fill pointer back			*/
   		ADDQ.L	#1,D0						/* Increment the byte count						*/
   		CMP.L		D0,D1						/* Compare the two byte counts					*/
   		BPL		@GoodBye					/* Don't worry if no overflow						*/
   		
   		MOVEA.L	EMPTY(A5),A2			/* Get the empty pointer							*/
			ADDA.L	#1,A2						/* Increment the MT pointer (bye old stuff!) */
   		CMPA.L	A2,A3						/* See if the buffer is at the end				*/
   		BPL		@Skip2					/* Don't worry if we're not at the end			*/
   		
   		LEA.L		BUFF(A5),A2				/* Reset the empty pointer to the beginning	*/

@Skip2   SUBQ.L	#1,D0						/* No more bytes allowed							*/
			MOVE.L	A2,EMPTY(A5)			/* Put the modified empty pointer back			*/
			LEA.L		DRERR(A4),A2			/* Get the driver error location					*/
			MOVE.W	#INBUFFOV,(A2)			/* Return a buffer overflow error				*/
	
@GoodBye	MOVE.L	D0,COUNT(A5)			/* Put the new byte count back					*/

@GitGone MOVEM.L	(SP)+,D0-D1/A2-A5		/* Restore the registers							*/
			MOVE.W	(SP)+,SR					/* Restore the original status					*/
			RTS									/* Git already											*/
      };
	};

/************************************************************************************/
/* This is the interrupt called routine that outputs a byte to the SCC.					*/
/************************************************************************************/

pascal void PutOne ()
	{
   asm
   	{
   		MOVE.W	SR,-(SP)					/* Save the previous status condition			*/
			ORI.W		#0x0300,SR				/* Step the interrupt priority up to 3			*/
   		MOVEM.L	D0-D1/A2-A5,-(SP)		/* Save the registers								*/
			   		
   		JSR		MyStash					/* Get the address of the private data			*/

   		MOVE.L	D0,A4						/* Address the data structure						*/
   		MOVE.L	(A4),A4					/* Takes two steps for generic 68000 code		*/
   	
   		MOVEA.L	OUTBUFF(A4),A5			/* Address the output buffer 						*/
   		MOVE.L	COUNT(A5),D0			/* Get the byte count								*/
   		BEQ		@GoodBye					/* Don't do anything if buffer is empty		*/

@SCCChk	MOVEA.L	SCCWRITE(A4),A1		/* Address port write control register	 		*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			MOVE.B	#0x00,(A1)				/* Address register 0								*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			MOVEA.L	SCCREAD(A4),A0			/* Address port read control register 0 		*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			CLR.L		D1							/* Just for s's and g's, we'll clear D1		*/
			MOVE.B	(A0),D1					/* Put the status byte into D1					*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			BTST		#0x02,D1					/* See if the TX buffer is empty first			*/
			BEQ		@GoodBye					/* Don't do it if the buffer's not MT			*/
			
   		MOVEA.L	EMPTY(A5),A2			/* Load the empty pointer							*/

      	MOVEA.L  SCDWRITE(A4),A1		/* Place register address into address reg	*/
      	MOVE.L   (SP),(SP)   			/* This kills settling time  						*/
			
@SCWrite MOVE.B	(A2)+,(A1)				/* Output the byte to the SCC						*/
      	
   		LEA.L		BUFF(A5),A3				/* Get the beginning of the buffer				*/
   		MOVE.L	SIZE(A5),D1				/* Get the buffer size								*/
   		ADDA.L	D1,A3						/* Point to the end of the buffer				*/
   		SUBA.L	#1,A3						/* Take 1 off to include first byte				*/

   		CMPA.L	A2,A3						/* See if the buffer is at the end				*/
   		BPL		@Skip						/* Don't worry if we're not at the end			*/
   		   		
   		LEA.L		BUFF(A5),A2				/* Reset the empty pointer to the beginning	*/
   		
@Skip		MOVE.L	A2,EMPTY(A5)			/* Put the modified empty pointer back			*/
   		SUBQ.L	#1,D0						/* Decrement the byte count						*/
	
@GoodBye	MOVE.L	D0,COUNT(A5)			/* Put the new byte count back					*/
			BNE		@GitNow					/* Reset the interrupt only if MT				*/
			
			MOVEA.L	SCCWRITE(A4),A1		/* Address port B write control register	 	*/
      	MOVE.B   #RESETTBE,(A1)			/* Reset the TBE interrupt							*/
			
@GitNow	MOVEM.L	(SP)+,D0-D1/A2-A5		/* Restore the registers							*/
			MOVE.W	(SP)+,SR					/* Restore the original status					*/
			RTS									/* Git already											*/
      };
	};
		   
/************************************************************************************/
