/************************************************************************************/
/* 								Low Level MIDI Interface Driver									*/
/************************************************************************************/
/* Version: 1.0			Written by Chris Marshall September 22nd, 1989					*/
/* 										1989 by Chris Marshall										*/
/* Written using the THINK C version 4.0 development system.								*/
/************************************************************************************/
/* These are driver-exclusive declarations for the MIDI driver.                     */
/************************************************************************************/

/************************************************************************************/
/* Default settings																						*/
/************************************************************************************/

#define	DEFCLOCK		CL1MHZ	/* Default clock index											*/

/************************************************************************************/
/* Miscellaneous driver constants																	*/
/************************************************************************************/

#define	DRVROPEN		0			/* Open call for driver											*/
#define	DRVRPRIME	1			/* Prime call for driver										*/
#define	DRVRCONTROL	2			/* Control call for driver										*/
#define	DRVRSTATUS	3			/* Status call for driver										*/
#define	DRVRCLOSE	4			/* Close call for driver										*/

/************************************************************************************/
/* These are port setup constants.                                                  */
/************************************************************************************/

#ifndef NULL					/* see if NULL has already been done.							*/
	#define	NULL	0L			/* Define it as 0L.													*/
#endif

#define	ZERO		0x00		/* Self documenting.													*/

#define	REG4		0x04  	/* 1 stop bit/char.													*/
#define	REG1		ZERO  	/* Interrupts disabled for now.									*/
#define	REG2		ZERO		/* No vector.															*/
#define	REG3		0xC0  	/* Rx: 8 bits/char, and disable Rx. 							*/
#define	REG5		0x62  	/* Tx: 8 bits/char, RTS enable, disable Tx.					*/
#define	REG6		ZERO		/* No sync stuff														*/
#define	REG7		ZERO		/* No sync stuff														*/
#define	REG10		ZERO		/* We're just simple two-wire async - no fancy stuff.		*/
#define	REG11		0x28  	/* Set clock source to TRxC pin for Tx and Rx.				*/
#define	REG12		ZERO		/* No time constant - external clock source.					*/
#define	REG13		ZERO		/* No time constant - external clock source.					*/
#define	REG14		ZERO		/* No baud rate generator either.								*/

#define	REG3B		0xC1  	/* Rx: 8 bits/char, and enable Rx. 								*/
#define	REG5B		0x6A  	/* Tx: 8 bits/char, RTS enable, Tx enable.					*/

#define	REG15		0x08  	/* Reset DCD interrupt for the mouse.							*/
#define	RESETTBE	0x28		/* Reset the Tx pending interrupt.								*/
#define	RESETEXT	0x10  	/* Reset EXT interrupt.                                	*/
#define	REG1B		0x13  	/* Interrupt register to Rx every char with Tx.				*/
#define	REG9		0x0A  	/* Master interrupt enable, and Rx/Tx interrupt enable.	*/

#define	EXTONLY	0x01		/* Enable only the EXT interrupt									*/

/************************************************************************************/
/* Address offsets for the port control registers and interrupt vectors.				*/
/************************************************************************************/

#define	TBEOFFSET	DEFPORT * 4					/* Offset to the port base					*/
#define	CAVOFFSET	TBEOFFSET + 2				/* Long word offsets into Lvl2DT table	*/

#define	CTLOF			(long) (2 * DEFPORT)		/* Byte offsets to the SCC	(control)	*/
#define	DATOF			(long) (4 + CTLOF)		/* Data offset									*/

/************************************************************************************/
/* Equates for assembly language																		*/
/************************************************************************************/

/**** These are offset constants that reflect the data structures defined below  ****/

/* Buffer structure */

#define	FILL			0
#define	EMPTY			4
#define	COUNT			8
#define	SIZE			12
#define	BUFF			16

/* Data structure */

#define	INBUFF		0
#define	OUTBUFF		4
#define	OLDVECT1		8
#define	OLDVECT2		12
#define	UNUSED		16
#define	SCDREAD		20
#define	SCDWRITE		24
#define	SCCREAD		28
#define	SCCWRITE		32
#define	CLOCKINDEX	36
#define	DRERR			38

/************************************************************************************/
/* Universal type definitions																			*/
/************************************************************************************/

typedef struct	MyBuffer			/* The structure of a buffer							 		*/
	{																					/* Byte Offset 	*/
	unsigned char	*fill;		/* Fill pointer										0			*/
	unsigned char	*empty;		/* Empty pointer										4			*/
	long				count;		/* Valid content byte count						8			*/
	long				buffSize;	/* Allocated buffer length							12			*/
	unsigned char	buff [1];	/* Start of allocated buffer						16			*/
	} MyBuffer, *MyBufferPtr;
	
typedef struct MyData			/* The driver's "private data" structure					*/
	{																					/* Byte Offset 	*/
	MyBufferPtr	inputBuffer;	/* Input buffer										0			*/
	MyBufferPtr	outputBuffer;	/* Output buffer										4			*/
	Ptr				oldVector1;	/* Original Character Available vector			8			*/
	Ptr				oldVector2;	/* Original Transmit Buffer Empty vector		12			*/
	long				unUsed;		/* Scratch (not used yet)							16			*/
	unsigned char	*sccDataRd;	/* SCC data read address							20			*/
	unsigned char	*sccDataWr;	/* SCC data write address							24			*/
	unsigned char	*sccCtrlRd;	/* SCC control read address						28			*/
	unsigned char	*sccCtrlWr;	/* SCC control write address						32			*/
	int				myClock;		/* The clock source index							36			*/
	int				driverErr;	/* Error return from interrupt calls			38			*/
	} MyData, *MyDataPtr;

/************************************************************************************/
/* Function prototypes																					*/
/************************************************************************************/

/* These are for the "MIDI Driver.C" file */

extern int 	OpenMyDriver (DCtlEntry *);
extern int 	MyDriverPrime (HParamBlockRec *);
extern int 	MyControl (CntrlParam *);
extern int 	MyStatus (CntrlParam *);
extern int 	CloseMyDriver (void);
extern int 	SetSize (long);
extern int 	ClearBuffers (void);
extern int 	ReadData (HParamBlockRec *);
extern int 	WriteData (HParamBlockRec *);
extern long	ReadBuff (unsigned char *, long);
extern long WriteBuff (unsigned char *, long);
extern int	ReplaceVectors (void);
extern int 	RestoreVectors (void);
extern int	SetUpRegisters (void);
extern int	ResetRegisters (void);
extern int	AddressSCC (void);
extern int	DoRegisters (int);
extern int	RestoreRegisters (void);

/* These are for the "Nasty Assembly Stuff.C" file */

extern MyDataPtr		*MyStash (void);
extern int				WriteCtlReg (unsigned char, unsigned char);
extern unsigned char	ReadCtlReg (unsigned char);
extern pascal void	GetOne (void);
extern pascal void	PutOne (void);

/************************************************************************************/
