/*

	Created: Thursday, March 30, 1989 at 4:48 PM
    Offscreen.Think.h
    C Interface to the Macintosh Libraries

    Copyright Apple Computer, Inc.  1985-1989
    All rights reserved

	History:
	=======
	DAS	14-Jul-89	began port of MPW interface to Think C 4.0 version
	DAS	04-Feb-90	more of same; added 3 missing definitions; some reformatting
	DAS	12-Feb-90	added some missing definitions
	DAS	30-Mar-90	added missing Palette Manager glue routines
	DAS	05-Apr-90	added missing ANSI C glue
	DAS	14-Apr-90	added new palette usage constants
	DAS	17-Apr-90	updated to current MPW versions of <QDOffscreen.h> and <Palettes.h>
	DAS	20-May-90	updated to current MPW versions of <QDOffscreen.h> and <Palettes.h>
	DAS	06-Jun-90	added more IM Chap VI inlines
	DAS	14-Jun-90	fixed bugs related to "gwflags" short enum declaration
*/


#ifndef _H_Offscreen_Think
#define _H_Offscreen_Think

#ifdef THINK_C

/* Alert: Think 4.0 is missing certain C++ and ANSI C language features! */
#ifndef _H_CPlusCovers
#define class		struct
#endif _H_CPlusCovers

#ifndef const
#define const		/**/
#endif const


#include <QuickDraw.h>
#include <Color.h>

/* Alert: CQD is missing these */
#define ditheredCopy			64
#define maskModeBit				0			/* BSAlert: wrong value here!!! */

extern RGBColor		HiliteRGB: 0xDA0;		/* System global variable */

#endif THINK_C


/*
**	use NGetTrapAddress with this constant to detect presence of QD-32
**	and other related traps
*/
#define TRAPID_BlessedQD32		((unsigned short)0xAB03)
#define TRAPID_BlessedPltt32	((unsigned short)0xAAA2)
#define TRAPID_BitMapToRgn		((unsigned short)0xA8D7)


/* New QDError codes */
#define cRgnTooComplex			((QDError)-147)	/* sizeof rgn > 64K error */
#define pixMapTooBigErr			((QDError)-148)	/*  */
#define cNotEnoughStack			((QDError)-149)	/* no room at the inn */
#define cMatchErr				((QDError)-150)	/*  */
#define cTempMemErr				((QDError)-151)	/*  */
#define cNoMemErr				((QDError)-152)	/*  */
#define cRangeErr				((QDError)-153)	/*  */
#define cProtectErr				((QDError)-154)	/*  */
#define cDevErr					((QDError)-155)	/*  */
#define cResErr					((QDError)-156)	/*  */
#define cDepthErr				((QDError)-157)	/* invalid pixel depth */
#define cParmErr				((QDError)-158)	/*  */
#define cBitMapToRgnErr			((QDError)-500)	/* sizeof rgn > 64K error */

/* Flag bits passed to NewGWorld */
enum
{
	pixPurgeBit			= 0,		/* make the offscreen buffer purgeable */
	noNewDeviceBit		= 1,		/* don't create an offscreen device */
	pixelsPurgeableBit	= 6,		/* pixels are purgeable */
	pixelsLockedBit		= 7,		/* pixels are locked */
	mapPixBit			= 16,		/* set if color table mapping occurred */
	newDepthBit			= 17,		/* set if pixels were scaled to a different depth */
    alignPixBit			= 18,		/* set if pixels were realigned to screen alignment */
	newRowBytesBit		= 19,		/* set if pixmap was reconfigured in a new rowBytes */
	reallocPixBit		= 20,		/* set if offscreen buffer had to be reallocated */
	clipPixBit			= 28,		/* set if pixels were or are to be clipped */
	stretchPixBit		= 29,		/* set if pixels were or are to be stretched/shrinked */
	ditherPixBit		= 30,		/* set if pixels were or are to be dithered */
	gwFlagErrBit		= 31
};

typedef unsigned long GWorldFlags;

#define gwNoFlags		((GWorldFlags)0)
#define pixPurge		((GWorldFlags)(1<<pixPurgeBit))
#define noNewDevice		((GWorldFlags)(1<<noNewDeviceBit))
#define pixelsPurgeable	((GWorldFlags)(1<<pixelsPurgeableBit))
#define pixelsLocked	((GWorldFlags)(1<<pixelsLockedBit))
#define mapPix			((GWorldFlags)(1<<mapPixBit))
#define newDepth		((GWorldFlags)(1<<newDepthBit))
#define alignPix		((GWorldFlags)(1<<alignPixBit))
#define newRowBytes		((GWorldFlags)(1<<newRowBytesBit))
#define reallocPix		((GWorldFlags)(1<<reallocPixBit))
#define clipPix			((GWorldFlags)(1<<clipPixBit))
#define stretchPix		((GWorldFlags)(1<<stretchPixBit))
#define ditherPix		((GWorldFlags)(1<<ditherPixBit))
#define gwFlagErr		((GWorldFlags)(1<<gwFlagErrBit))


/* Mask equivalents of the above bits  */
#define pixelsPurgeableMask		0x00000040L
#define pixelsLockedMask		0x00000080L
#define mapPixMask				0x00010000L
#define newDepthMask			0x00020000L
#define alignPixMask			0x00040000L
#define newRowBytesMask			0x00080000L
#define reallocPixMask			0x00100000L
#define clipPixMask				0x10000000L
#define stretchPixMask			0x20000000L
#define ditherPixMask			0x40000000L


/* Type definition of a GWorldPtr */
typedef CGrafPtr	GWorldPtr;

/* new palette usage constants */
#define pmInhibitG2				((unsigned short)256)
#define pmInhibitC2				((unsigned short)512)
#define pmInhibitG4				((unsigned short)1024)
#define pmInhibitC4				((unsigned short)2048)
#define pmInhibitG8				((unsigned short)4096)
#define pmInhibitC8				((unsigned short)8192)


#ifdef __safe_link
extern "C" {
#endif



pascal QDErr NewGWorld(
	GWorldPtr	*offscreenGWorld,
	short		pixelDepth,
	const Rect	*boundsRect,
	CTabHandle	cTable,
	GDHandle	aGDevice,
	long		flags )
= { 0x203C, 0x0016, 0x0000, 0xAB1D };

pascal Boolean LockPixels(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x0001, 0xAB1D };

pascal void UnlockPixels(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x0002, 0xAB1D };

pascal long UpdateGWorld(
	GWorldPtr	*offscreenGWorld,
	short		pixelDepth,
	const Rect	*boundsRect,
	CTabHandle	cTable,
	GDHandle	aGDevice,
	long		flags )
= { 0x203C, 0x0016, 0x0003, 0xAB1D };

pascal void DisposeGWorld(
	GWorldPtr	offscreenGWorld )
= { 0x203C, 0x0004, 0x0004, 0xAB1D };

pascal void GetGWorld(
	CGrafPtr	*port,
	GDHandle	*gdh )
= { 0x203C, 0x0008, 0x0005, 0xAB1D };

pascal void SetGWorld(
	CGrafPtr	port,
	GDHandle	gdh )
= { 0x203C, 0x0008, 0x0006, 0xAB1D };

pascal void CTabChanged(
	CTabHandle	ctab )
= { 0x203C, 0x0004, 0x0007, 0xAB1D };

pascal void PixPatChanged(
	PixPatHandle	ppat )
= { 0x203C, 0x0004, 0x0008, 0xAB1D };

pascal void PortChanged(
	GrafPtr		port )
= { 0x203C, 0x0004, 0x0009, 0xAB1D };

pascal void GDeviceChanged(
	GDHandle	gdh )
= { 0x203C, 0x000A, 0x0007, 0xAB1D };

pascal void AllowPurgePixels(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x000B, 0xAB1D };

pascal void NoPurgePixels(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x000C, 0xAB1D };

pascal long GetPixelsState(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x000D, 0xAB1D };

pascal void SetPixelsState(
	PixMapHandle	pm,
	long			state )
= { 0x203C, 0x0008, 0x000E, 0xAB1D };

pascal Ptr GetPixBaseAddr(
	PixMapHandle	pm )
= { 0x203C, 0x0004, 0x000F, 0xAB1D };

pascal QDErr NewScreenBuffer(
	const Rect		*globalRect,
	Boolean			purgeable,
	GDHandle		*gdh,
	PixMapHandle	*offscreenPixMap )
= { 0x203C, 0x000E, 0x0010, 0xAB1D };

pascal void DisposeScreenBuffer(
	PixMapHandle	offscreenPixMap )
= { 0x203C, 0x0004, 0x0011, 0xAB1D };

pascal GDHandle GetGWorldDevice(
	GWorldPtr		offscreenGWorld )
= { 0x203C, 0x0004, 0x0012, 0xAB1D };

pascal Boolean QDDone(
	GrafPtr			port )
= { 0x203C, 0x0004, 0x0013, 0xAB1D };

pascal long OffscreenVersion( void )
= { 0x203C, 0x0000, 0x0014, 0xAB1D };

pascal QDErr NewTempScreenBuffer(
	Rect			*globalRect,
	Boolean			purgeable,
	GDHandle		*gdh,
	PixMapHandle	*offscreenPixMap )
= { 0x203C, 0x000E, 0x0015, 0xAB1D };

pascal Boolean Pixmap32Bit(
	PixMapHandle pm)
= { 0x203C, 0x0004, 0x0016, 0xAB1D };

pascal PixMapHandle GetGWorldPixMap(
	GWorldPtr offscreenGWorld)
= { 0x203C, 0x0004, 0x0017, 0xAB1D };


/* may be contained in a later version of offscreen.h */
pascal QDErr BitMapToRegion(
	RgnHandle	region,
	BitMap		*bMap )
= { 0xA8D7 };


/* contained in MPWs <Palettes.h> */
pascal void CopyPalette(
	PaletteHandle	srcPalette,
	PaletteHandle	dstPalette,
	short			srcEntry,
	short			dstEntry,
	short			dstLength )
= { 0xAAA1 };

pascal long Entry2Index(
	short	entry )
= { 0x7000, 0xAAA2 };

pascal void RestoreDeviceClut(
	GDHandle	gdh )
= { 0x7002, 0xAAA2 };

pascal void ResizePalette(
	PaletteHandle	srcPalette,
	short			size )
= { 0x7003, 0xAAA2 };

pascal void SaveFore(
	ColorSpec	*theCSpec )
= { 0x700D, 0xAAA2 };

pascal void SaveBack(
	ColorSpec	*theCSpec )
= { 0x700E, 0xAAA2 };

pascal void RestoreFore(
	ColorSpec	*theCSpec )
= { 0x700F, 0xAAA2 };

pascal void RestoreBack(
	ColorSpec	*theCSpec )
= { 0x7010, 0xAAA2 };

pascal OSErr SetDepth(
	GDHandle	gdh,
	short		depth,
	short		whichFlags,
	short		flags )
= { 0x7013, 0xAAA2 };

pascal short HasDepth(
	GDHandle	gdh,
	short		depth,
	short		whichFlags,
	short		flags )
= { 0x7014, 0xAAA2 };

#ifdef __safe_link
}
#endif


#endif _H_Offscreen_Think
